let rec take i = function
  | [] -> []
  | x::xs ->
     if i > 0 then x :: take (i - 1) xs
     else [];;

let rec drop i = function
  | [] -> []
  | x::xs ->
     if i > 0 then drop (i-1) xs
     else x::xs;;

let rec nth n l =
  match n, l with
  | _, [] -> failwith "Empty list"
  | 0, x::_ -> x
  | _, _::xs -> nth (n-1) xs;;

type cell = Val of int | Empty;;
type board = cell list list;;

let rec is_perm_aux = function
  [], _ -> true
| Empty::xs, _ -> is_perm_aux (xs, Empty)
| v::xs, p -> (p <> v) && is_perm_aux (xs, v);; 
let is_perm x = is_perm_aux (List.sort compare x, Empty);;

let check_rows board = List.for_all is_perm board;;

let get_cols board = List.map (fun x -> List.map (nth x) board) [0;1;2;3];;
let check_cols board = List.for_all is_perm (get_cols board);;

let get board i j = nth i (nth j board);;

(* Returns the square with a left top-most corner at (i, j) *)
let get_square board i j = List.map (fun (dx, dy) -> get board (i+dx) (j+dy)) [(0,0);(1,0);(0,1);(1,1)];;
(* Returns all possible squares. *)
let get_squares board = List.map (fun (x, y) -> get_square board x y) [(0,0);(2,0);(0,2);(2,2)];;

let check_squares board = List.for_all is_perm (get_squares board);;

(* Returns if the board is valid. *)
let is_valid board = check_rows board && check_cols board && check_squares board;;

(* Counts the number of empty cells in the board *)
let count_empty board = List.length (List.filter (fun x -> x = Empty) (List.flatten board));;

let rec generate_all_seqs len = 
  if len = 0 then [[]]
  else let smaller_seqs = generate_all_seqs (len - 1) in
     List.flatten (List.map (fun x -> List.map (fun y -> x::y) smaller_seqs) [1; 2; 3; 4]);;  

(* Replaces as many Empty as there are. *)
let rec replace_in_row rr xx = match xx, rr with
| [], _ -> []
| Empty::xs, r::rem -> (Val r)::replace_in_row rem xs
| x::xs,_ -> x::replace_in_row rr xs;;

(* Undoes List.flatten, creating the 4x4 board. *)
let rec unflatten = function
 [] -> []
|xs -> let pref = take 4 xs and suf = drop 4 xs in
     pref::unflatten suf;;

let rec replace_in_board board assignments = unflatten (replace_in_row assignments (List.flatten board));;

let rec exists p = function
  [] -> None
| (x::xs) -> if p x then Some x else exists p xs;;


let find_solution board = 
   match exists (fun assigment -> is_valid (replace_in_board board assigment)) (generate_all_seqs (count_empty board)) with
     Some x -> Some (unflatten (replace_in_row x (List.flatten board)))
   | _ -> None;;

find_solution [ 
  [Empty; Empty; Val 2; Val 3];
  [Val 3; Empty; Empty; Val 4];
  [Val 2; Empty; Val 3; Empty];
  [Val 1; Empty; Empty; Empty]] =
Some [
 [Val 4; Val 1; Val 2; Val 3]; 
 [Val 3; Val 2; Val 1; Val 4];
 [Val 2; Val 4; Val 3; Val 1];
 [Val 1; Val 3; Val 4; Val 2]];;

find_solution [ 
  [Empty; Empty; Empty; Val 3];
  [Val 2; Val 3; Val 4; Empty];
  [Val 4; Empty; Val 3; Empty];
  [Val 3; Empty; Val 1; Empty]] =
Some [
  [Val 1; Val 4; Val 2; Val 3];
  [Val 2; Val 3; Val 4; Val 1];
  [Val 4; Val 1; Val 3; Val 2];
  [Val 3; Val 2; Val 1; Val 4]];;

find_solution [ 
  [Empty; Val 4; Empty; Empty];
  [Empty; Val 3; Empty; Val 2];
  [Empty; Empty; Empty; Val 4];
  [Empty; Val 2; Empty; Val 3]] =
Some
 [[Val 2; Val 4; Val 3; Val 1]; 
  [Val 1; Val 3; Val 4; Val 2];
  [Val 3; Val 1; Val 2; Val 4];
  [Val 4; Val 2; Val 1; Val 3]];;
