let rec take i = function
  | [] -> []
  | x::xs ->
     if i > 0 then x :: take (i - 1) xs
     else [];;

let rec drop i = function
  | [] -> []
  | x::xs ->
     if i > 0 then drop (i-1) xs
     else x::xs;;

let rec nth n l =
  match n, l with
  | _, [] -> failwith "Empty list"
  | 0, x::_ -> x
  | _, _::xs -> nth (n-1) xs;;

type cell = Val of int | Empty;;
type board = cell list list;;

let rec check_perm_aux = function
  [], _ -> true
| Empty::xs, _ -> check_perm_aux (xs, Empty)
| v::xs, p -> (p <> v) && check_perm_aux (xs, v);; 
let check_perm x = check_perm_aux (List.sort compare x, Empty);;

let check_rows board = List.for_all check_perm board;;

let get_cols board = 
   List.map (fun x -> List.map (nth x) board) [0;1;2;3;4;5;6;7;8];;

let check_cols board = List.for_all check_perm (get_cols board);;

let get board i j = nth i (nth j board);;

(* Retuns the square with a left top-most corner at (i, j) *)
let get_square board i j = List.map (fun (dx, dy) -> get board (i+dx) (j+dy)) [(0,0);(1,0);(0,1);(1,1);(2,0);(0,2);(2,1);(1,2);(2,2)];;

let get_squares board = List.map (fun (x, y) -> get_square board x y) [(0,0);(3,0);(0,3);(3,3);(6,0);(6,3);(3,6);(0,6);(6,6)];;

let check_squares board = List.for_all check_perm (get_squares board);;

let check_board board = check_rows board && check_cols board && check_squares board;;

let count_empty board = List.length (List.filter (fun x -> x = Empty) (List.flatten board));;

type 'a option = None | Some of 'a;;

(* Replaces the first occurrence of an empty. *)
let rec replace_first_in_row v xx = match xx with
| [] -> []
| Empty::xs -> (Val v)::xs
| x::xs-> x::replace_first_in_row v xs;;

let rec unflatten = function
 [] -> []
|xs -> let pref = take 9 xs and suf = drop 9 xs in
     pref::unflatten suf;;

let rec exists p = function
  [] -> None
| (x::xs) -> if p x then Some x else exists p xs;;

let replace_first_in_board board v = unflatten (replace_first_in_row v (List.flatten board));;

let rec explore board cnt = 
   if cnt = 0 then Some board
   else let rec explore_options cur =
      if cur = 10 then None
      else let new_board = replace_first_in_board board cur in
         if check_board new_board then 
            let ans = explore new_board (cnt-1) in 
            if ans = None then explore_options (cur+1)
            else ans
         else explore_options (cur+1)
      in explore_options 1;;


let find_solution board = explore board (count_empty board);;

find_solution [ 
  [Empty; Val 6; Empty; Empty; Empty; Val 9; Val 3; Empty; Empty];
  [Val 3; Val 5; Val 7; Val 8; Empty; Empty; Empty; Val 1; Empty];
  [Empty; Val 2; Empty; Empty; Val 5; Empty; Empty; Empty; Empty];
  [Val 6; Empty; Val 1; Empty; Empty; Empty; Val 7; Empty; Empty];
  [Val 4; Empty; Empty; Empty; Val 3; Val 7; Empty; Empty; Val 8];
  [Empty; Empty; Val 3; Empty; Empty; Empty; Val 1; Empty; Val 2];
  [Val 7; Empty; Val 9; Empty; Empty; Val 5; Val 8; Empty; Val 6];
  [Empty; Empty; Empty; Empty; Empty; Val 3; Val 9; Empty; Empty];
  [Val 8; Empty; Empty; Val 7; Empty; Empty; Val 2; Empty; Val 1];
  ] = 
Some
 [[Val 1; Val 6; Val 8; Val 4; Val 7; Val 9; Val 3; Val 2; Val 5];
  [Val 3; Val 5; Val 7; Val 8; Val 2; Val 6; Val 4; Val 1; Val 9];
  [Val 9; Val 2; Val 4; Val 3; Val 5; Val 1; Val 6; Val 8; Val 7];
  [Val 6; Val 8; Val 1; Val 5; Val 4; Val 2; Val 7; Val 9; Val 3];
  [Val 4; Val 9; Val 2; Val 1; Val 3; Val 7; Val 5; Val 6; Val 8];
  [Val 5; Val 7; Val 3; Val 9; Val 6; Val 8; Val 1; Val 4; Val 2];
  [Val 7; Val 4; Val 9; Val 2; Val 1; Val 5; Val 8; Val 3; Val 6];
  [Val 2; Val 1; Val 5; Val 6; Val 8; Val 3; Val 9; Val 7; Val 4];
  [Val 8; Val 3; Val 6; Val 7; Val 9; Val 4; Val 2; Val 5; Val 1]];;



find_solution [ 
  [Empty; Empty; Val 2; Val 3];
  [Val 3; Empty; Empty; Val 4];
  [Val 2; Empty; Val 3; Empty];
  [Val 1; Empty; Empty; Empty]] =
Some [
 [Val 4; Val 1; Val 2; Val 3]; 
 [Val 3; Val 2; Val 1; Val 4];
 [Val 2; Val 4; Val 3; Val 1];
 [Val 1; Val 3; Val 4; Val 2]];;

find_solution [ 
  [Empty; Empty; Empty; Val 3];
  [Val 2; Val 3; Val 4; Empty];
  [Val 4; Empty; Val 3; Empty];
  [Val 3; Empty; Val 1; Empty]] =
Some [
  [Val 1; Val 4; Val 2; Val 3];
  [Val 2; Val 3; Val 4; Val 1];
  [Val 4; Val 1; Val 3; Val 2];
  [Val 3; Val 2; Val 1; Val 4]];;



